#!/usr/bin/env python
#   -*- encoding: UTF-8 -*-
#
#   :Author: tundish
#   :Date: 2007-02-14
#   :Copyright: 2007 Thuswise Ltd
#   :Description:   sprinkle is a Python stomp broker
#                   coded for Python 2.3.4 and above
#   :Contact: tundish at thuswise . org . uk
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#   
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>

import sys
import os
import os.path
import signal
import tempfile
import shutil
import socket
import signal
import syslog
import random
from syslog import syslog as log
from syslog import openlog, closelog
from optparse import OptionParser
from ConfigParser import SafeConfigParser
from ConfigParser import NoSectionError, DuplicateSectionError

DEFAULT_PORT = 12000
DEFAULT_ATTENTION = 0.3

class SimpleStompFrame(object):
    """ A base class for Stomp frames """

    def __init__(self, t, msg, **kwargs):
        self.type = t
        self.msg = msg
        self.hdr = kwargs.copy()

    def __repr__(self):
        keys = self.hdr.keys()
        keys.sort()
        return "%s\n%s\n\n%s\x00" % (
        self.type.upper(),
        '\n'.join(["%s: %s" % (k,self.hdr[k])for k in keys]),
        self.msg) 

    def __str__(self):
        return repr(self)

    def requestReceipt(self, rcpt):
        self.hdr["receipt"] = rcpt
        return self

    def joinTransaction(self, trct):
        self.hdr["transaction"] = trct
        return self

class ContentedStompFrame(SimpleStompFrame):
    """ Stomp frames with content length"""

    def __init__(self, t, msg, **kwargs):
        SimpleStompFrame.__init__(self, t, msg, **kwargs)
        self.hdr["content-length"] = str(len(msg))

def connect(login, passcode):
    if not login or not passcode:
        return None
    else:
        return SimpleStompFrame("connect", "",
        login=login, passcode=passcode)

def connected(session):
    if not session:
        return None
    else:
        return SimpleStompFrame("connected", "",
        session=session)

def subscribe(destination, ack="auto"):
    if not destination or ack not in ("client", "auto"):
        return None
    else:
        return SimpleStompFrame("subscribe", "",
        destination=destination, ack=ack)

def unsubscribe(destination):
    if not destination:
        return None
    else:
        return SimpleStompFrame("unsubscribe", "",
        destination=destination)

def receipt(id):
    if not id:
        return None
    else:
        rv = SimpleStompFrame("receipt", "")
        rv.hdr["receipt-id"] = id
        return rv

def begin(transaction):
    if not transaction:
        return None
    else:
        return SimpleStompFrame("begin", "",
        transaction=transaction)

def commit(transaction):
    if not transaction:
        return None
    else:
        return SimpleStompFrame("commit", "",
        transaction=transaction)

def abort(transaction):
    if not transaction:
        return None
    else:
        return SimpleStompFrame("abort", "",
        transaction=transaction)

def message(msg, destination, id):
    if not destination or not id:
        return None
    else:
        rv = ContentedStompFrame("message", msg,
        destination=destination)
        rv.hdr["message-id"] = id
        return rv

def ack(id):
    if not id:
        return None
    else:
        rv = SimpleStompFrame("ack", "")
        rv.hdr["message-id"] = id
        return rv

def send(msg, destination):
    if not destination:
        return None
    else:
        return ContentedStompFrame("send", msg,
        destination=destination)

def error(body, message):
    return ContentedStompFrame("error", body, message=message)

def disconnect():
    return SimpleStompFrame("disconnect", "")

class ParseError(Exception):
    
    def __init__(self, arg):
        self.arg = arg

    def __str__(self):
        return repr(self.arg)

class Deframer(object):
    """ A processor for Stomp strings """

    def __init__(self, buf = []):
        self.buf = buf

    def __iter__(self):
        return self

    def load(self, data):
        if str(data):
            self.buf.append(data)

    def next(self):
        work = self.buf[:]  # a working copy of the buffer
        data = ""           # the string to be processed
        lines = None        # a list for line-separated data
        
        while True:
            try:
                data += work.pop(0)
                lines = data.splitlines()
                blankPos = lines.index('')
            except ValueError:
                # no end of header - need more data
                continue
            except IndexError:
                # nothing more in working buffer
                raise StopIteration

            # data contains a full header
            # grab the frame type
            t = lines[0]
            try:
                hdr = dict([(k.strip(),v.strip()) for k,v in
                [l.split(':') for l in lines[1:blankPos]]])
            except ValueError:
                # header contains at least one line which is
                # not a key:value pair - discard it
                log(syslog.LOG_NOTICE,"malformed header")
                data = ""
                del self.buf[0:len(self.buf) - len(work)]
                continue
            except IndexError:
                # nothing beyond the header - need more data
                continue

            # we have a type and a header - now read the message
            data = '\n'.join(lines[blankPos + 1:])
            try:
                # read up to content-length
                cLen = int(hdr["content-length"])
                while len(data) <= cLen:
                    data += work.pop(0)
            except ValueError:
                # content-length not an integer - discard header
                log(syslog.LOG_NOTICE,"invalid content-length")
                data = ""
                del self.buf[0:len(self.buf) - len(work)]
                continue
            except IndexError:
                # nothing more in working buffer
                raise StopIteration
            except KeyError:
                # no content-length
                cLen = 0

            # the next null after content-length
            # is the end of the message
            nullPos = data.find('\x00', cLen)
            while nullPos == -1:
                try:
                    data += work.pop(0)
                    nullPos = data.find('\x00', cLen)
                except IndexError:
                    # nothing more in working buffer
                    raise StopIteration

            msg = data[:nullPos]
            # replace the remainder back in the buffer
            try:
                del self.buf[0:len(self.buf) - len(work)]
                rmdr = data[nullPos + 1:]
                if rmdr.strip():
                    self.buf.insert(0,rmdr)
            except IndexError:
                pass
            
            return t, hdr, msg

class MessageManager(object):

    """ A class which accepts submissions to message queues """
    
    def __init__(self, home="queues"):
        self.root = os.path.join(os.path.dirname(__file__), home)

    def _msgRef(self, q):
        """ Return a file handle and path for the message q """
        dir = os.path.join(self.root, q)
        return tempfile.mkstemp(".msg", q + '-', dir, text = True)

    def queues(self):
        try:
            return os.listdir(self.root)
        except OSError:
            log(syslog.LOG_ALERT, "No queue directory")
            raise

    def messages(self):
        """ Iterates over msgs. Queues will appear oldest msg first"""
        for q in self.queues():
            qPath = os.path.join(self.root,q)
            msgs = [os.path.join(qPath,i) for i in os.listdir(qPath)]
            timed = [(os.path.getctime(i),i) for i in msgs]
            timed.sort()
            for t,m in timed:
                yield m

    def msgInfo(self, ref):
        """ Return the queue and msg id from the msg reference. May
        throw ValueError if the message reference does not exist"""
        b = os.path.basename(ref)
        sep = b.index('-')
        return (b[:sep], b[sep + 1:b.index('.')])

    def submit(self, q, content):
        """ Submit a msg. Returns the msg id if successful """
        try:
            fd, path = self._msgRef(q)
        except OSError:
            return ""
        os.write(fd,content)
        os.close(fd)
        return self.msgInfo(path)[1]

class UserManager(object):

    """ A class which manages a user's view of his current messages """
    
    def __init__(self, home="users"):
        self.root = os.path.join(os.path.dirname(__file__), home)
        self.configs = {}

    def _userRef(self, u):
        return os.path.join(self.root, u, u + ".cfg")

    def _store(self, u):
        try:
            of = file(self._userRef(u),'w')
        except IOError:
            return False
        self.configs[u].write(of)
        of.close()
        return True

    def configure(self):
        """ Call this method after instantiation """
        for i in self.users():
            cp = SafeConfigParser()
            cp.read(self._userRef(i))
            self.configs[i] = cp

    def userPath(self, u, msg):
        """ This is where the message will end up for this user"""
        return os.path.join(self.root, u, os.path.basename(msg))

    def users(self):
        """ Returns a list of defined users """
        try:
            return [ i for i in os.listdir(self.root) if os.path.isdir(
            os.path.join(self.root,i))]
        except OSError:
            log(syslog.LOG_ALERT, "No users directory")
            raise
        
    def subscribers(self, q):
        """ Returns a list of users subscribed to this queue """
        return [k for k,v in self.configs.items() if v.has_section(q)]

    def check(self, user, passwd):
        """ Verify user login """
        try:
            cfg = self.configs[user]
            if cfg.defaults()["passcode"] == passwd:
                return True
            else:
                return False
        except KeyError:
            return False

    def total(self, u):
        """ Return the number of unclaimed messages for this user"""
        return len(os.listdir(os.path.join(self.root, u))) - 1

    def add(self, client, passwd, dest, ack):
        """ Add a user to a destination queue """
        try:
            cp = self.configs[client]
            if cp.defaults()["passcode"] != passwd:
                return False
            cp.add_section(dest)
        except KeyError:
            # no config or no password means missing or
            # invalid user setup
            # TODO: Flag this for attention
            return False
        except DuplicateSectionError:
            # client is already subscribed
            # but maybe the options are to be changed
            pass
        cp.set(dest,"ack", str(ack))
        return self._store(client)
    
    def remove(self, client, passwd, dest):
        """ Remove a user from a destination queue """
        try:
            cp = self.configs[client]
            if cp.defaults()["passcode"] != passwd:
                return False
            elif cp.remove_section(dest):
                return self._store(client)
            else:
                return False
        except KeyError:
            # no config means missing or invalid user setup
            # TODO: Flag this for attention
            return False
    
    def pending(self, u):
        """ Generate pending (queue, msg, content) for a user
        NB: if no ack is required, the message will be destroyed
        Raises KeyError if any details are invalid"""
        try:
            cF = os.path.basename(self._userRef(u))
            d = os.path.dirname(self._userRef(u))
            uc = self.configs[u]
            for fN in os.listdir(d):
                if fN != cF:
                    # derive values for queue, message id, content
                    q = fN.split('-')[0]
                    msg = os.path.splitext(fN)[0]
                    p = os.path.join(d,fN)
                    f = file(p)
                    c = f.read()
                    try:
                        # implement auto-acknowledgement
                        if uc.get(q,"ack") == "auto":
                            os.remove(p)
                    except NoSectionError:
                        # no longer subscribed to this queue
                        os.remove(p)
                        continue
                    yield (q,msg,c)
                    f.close()
        except OSError:
            raise StopIteration
        except:
            raise KeyError

    def acknowledge(self, u, id):
        """ Acknowledge the delivery of a message"""
        d = os.path.dirname(self._userRef(u))
        p = os.path.join(d,id)
        uc = self.configs[u]
        q = id.split('-')[0]
        if uc.get(q,"ack") == "client":
            try:
                os.remove(p)
                return True
            except OSError:
                pass
        return False
            
class ForkMachine(object):
    """
        This class can adopt the interfaces of two other
        classes using a mix-in technique. The fork method
        returns two different types of object, one in a
        parent process and the other in a child process
    """ 

    def __init__(self):
        """ Keep a record of our base classes when we were first made """
        self.bases = self.__class__.__bases__
        signal.signal(signal.SIGCHLD, self._reapChild)

    def _mixIn(self, mixInClass, args, kwargs):
        """ Dynamically adopt the interface of a mix-in class and
        instantiate ourself as a newly initialised object """
        self.__class__ = type(mixInClass.__name__,
        (mixInClass, self.__class__) + self.bases, {})
        mixInClass.__init__(self, *args, **kwargs)

    def _reapChild(self, signum, stackframe):
        """ This method removes zombie children. Override it for
        more specific behaviour """
        while True:
            try:
                result = os.waitpid(-1, os.WNOHANG)
                if not result[0]:
                    break
            except:
                break
        signal.signal(signal.SIGCHLD, self._reapChild)

    def fork(self, pClass, pArgs, pKwargs, cClass, cArgs, cKwargs):
        self.childId = os.fork()
        if self.childId:
            # I'm the parent
            if pClass:
                self._mixIn(pClass, pArgs, pKwargs)
        else:
            self._mixIn(cClass, cArgs, cKwargs)
        return self

    def isParent(self):
        return bool(self.childId)
        
class Spooler(ForkMachine):
    """
        TODO: Override _reapChild to fork a new Listener
    """
    hup = False

    def __init__(self, um, mm):
        """ This initialiser takes two objects: a User Manager,
        and a Message Manager"""
        ForkMachine.__init__(self)
        self.um = um
        self.mm = mm
        self.states = {
        "start": self.do_start,
        "spool": self.do_spool
        }
        log(syslog.LOG_INFO,"Starting up...")

    def __call__(self):
        state = "start"
        while True:
            try:
                state = self.states[state]()
            except KeyError:
                if self.opts.debug:
                    log(syslog.LOG_NOTICE,"end state: %s" % state)
                break

    def do_start(self):
        self.um.configure()
        return "spool"

    def do_spool(self):
        """ Visit all current messages and copy them out to their
        subscribers. Delete when done """
        if self.opts.allow_subscribe and self.hup:
            self.um.configure()
            self.hup = False
            if self.opts.debug:
                log(syslog.LOG_NOTICE,"Spooler reconfigured")
        for m in self.mm.messages():
            try:
                q, msg = self.mm.msgInfo(m)
            except ValueError:
                pass
            for s in self.um.subscribers(q):
                shutil.copy2(m, self.um.userPath(s,m))
            os.remove(m)
        return "spool"

# This process accepts the SIGHUP signal
def hupHandler(sig, frm):
    Spooler.hup = True

signal.signal(signal.SIGHUP, hupHandler)

class Listener(object):
    """
        TODO: Fork new handler processes
    """

    def __init__(self, id):
        self.spoolerPid = id
        self.states = {
        "start": self.do_listen,
        "accept": self.do_accept
        }

    def do_listen(self):
        """ Listen on a socket."""
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        self.sock.bind(("",self.opts.port))
        log(syslog.LOG_INFO,"Listening on port %d" % self.opts.port)
        self.sock.listen(1)
        return "accept"

    def do_accept(self):
        """Fork on connection"""
        while True:
            try:
                self.conn, self.addr = self.sock.accept()
            except KeyboardInterrupt:
                raise
            except:
                continue

            self._reapChild(None, None)

            try:
                self.fork(None, None, None, Worker, [self.spoolerPid], {})
            except:
                log(syslog.LOG_ERR,"Fork failed")
                self.conn.close()
                continue

            if self.isParent():
                # I'm still a Listener
                self.conn.close()
                continue
            else:
                # I'm now a Worker
                return "start"

class Worker(object):

    """ The object which handles a connection """

    def __init__(self, id):
        self.spoolerPid = id
        self.states = {
        "start": self.do_start,
        "connect": self.do_connect,
        "listen": self.do_listen,
        "deliver": self.do_deliver,
        "stop": self.do_stop
        }
        self.df = Deframer()
        self.transactions = {}

    def __call__(self):
        """ A hook to allow testing. Under normal operation,
        the parent __call__ method will loop """
        self.conn, self.addr = self.s.accept()
        state = "start"
        while True:
            try:
                state = self.states[state]()
            except KeyError:
                break

    def _recv(self,conn,addr):
        conn.setblocking(0)
        buf = ''
        try:
            # my telnet client sends "\x00\r\n"
            while '\x00' not in buf[-3:]:
                buf += conn.recv(1024)
            if self.opts.debug:
                log(syslog.LOG_NOTICE,"received: %s" % buf)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            # nothing to be read
            pass
        return buf

    def do_start(self):
        self.um.configure()
        return "connect"

    def do_connect(self):
        try:
            t, hdr, msg = self.df.next()
        except StopIteration:
            buf = self._recv(self.conn, self.addr)
            self.df.load(buf)
            return "connect"

        if t == "CONNECT":
            try:
                self.login = hdr["login"]
                self.passcode = hdr["passcode"]
            except KeyError:
                self.login = ""
                self.passcode = ""
                return "connect"
            if self.opts.debug:
                log(syslog.LOG_NOTICE,"client connected")
            if self.um.check(self.login, self.passcode):
                reply = connected(os.getpid())
                self.conn.sendall(str(reply))
                return "listen"
            else:
                reply = error(repr(hdr),"bad user credentials")
                self.conn.sendall(str(reply))
                return "stop"
        else:
            return "connect"
    
    def do_listen(self):
        # check existing frames
        try:
            t, hdr, msg = self.df.next()
            if self.opts.debug:
                log(syslog.LOG_NOTICE,t)
        except StopIteration:
            # decide whether to listen some more
            # or deliver messages instead
            if random.random() > self.opts.attention:
                return "deliver"
            else:
                buf = self._recv(self.conn, self.addr)
                self.df.load(buf)
                return "listen"

        if self.opts.debug:
            log(syslog.LOG_NOTICE,"checking: %s" % t)

        if t == "SUBSCRIBE":
            try:
                self.um.add(self.login, self.passcode
                , hdr["destination"], hdr["ack"])
                # tell the spooler the config has changed
                os.kill(self.spoolerPid, signal.SIGHUP)
            except KeyError:
                return "listen"
        elif t == "UNSUBSCRIBE":
            try:
                rv = self.um.remove(self.login, self.passcode
                , hdr["destination"])
                # tell the spooler the config has changed
                os.kill(self.spoolerPid, signal.SIGHUP)
                if not rv:
                    reply = error(msg,"failed to unsubscribe")
                    self.conn.sendall(str(reply))
            except KeyError:
                return "listen"
        elif t == "BEGIN":
            try:
                trx = hdr["transaction"]
                if not self.transactions.has_key(trx):
                    self.transactions[trx] = []
            except KeyError:
                return "listen"
        elif t == "ABORT":
            try:
                del self.transactions[hdr["transaction"]]
            except KeyError:
                return "listen"
        elif t == "SEND":
            try:
                trx = hdr["transaction"]
                self.transactions[trx].append((t, hdr, msg))
            except KeyError:
                # send message now (self.mm)
                if not self.mm.submit(hdr["destination"],msg):
                    reply = error(msg,"failed to send")
                    self.conn.sendall(str(reply))
        elif t == "COMMIT":
            try:
                trx = hdr["transaction"]
                for t, hdr, msg in self.transactions[trx]:
                    if t == "SEND":
                        # send message now (self.mm)
                        msgId = self.mm.submit(hdr["destination"],msg)
                    elif t == "ACK":
                        # acknowledge delivery (self.um)
                        msgId = self.um.acknowledge(self.login, hdr["message-id"])
                del self.transactions[trx]
            except KeyError:
                return "listen"
        elif t == "DISCONNECT":
            reply = error(t,"disconnection notice")
            self.conn.sendall(str(reply))
            return "stop"
                
        # send a receipt if necessary
        try:
            self.conn.sendall(str(receipt(hdr["receipt"])))
        except KeyError:
            # no receipt was requested
            pass

        return "listen"
            
    def do_deliver(self):
        """ Deliver existing messages """
        for q,msg,content in self.um.pending(self.login):
            self.conn.sendall(str(message(content,q,msg)))
        if random.random() < self.opts.attention:
            return "listen"
        else:
            return "deliver"
    
    def do_stop(self):
        self.conn.shutdown(2)
        self.conn.close()
        return "complete"

def parser():
    usage = """%prog [args]
    """
    p = OptionParser(usage)
    p.add_option("-p", "--port", type="int", default = DEFAULT_PORT,
    help = "the port number to listen on (default: %d)" % DEFAULT_PORT)
    p.add_option("-n", "--attention", type="float", default = DEFAULT_ATTENTION,
    metavar = "N", help = """how much (0.0 to 1.0) to listen to the client
(default: %0.2f)""" % DEFAULT_ATTENTION)
    p.add_option("-d", "--debug", action="store_true", default = False,
    help = "write debug messages to a log")
    p.add_option("--allow-subscribe", action="store_true",
    default = False, help = "allow users to subscribe to queues")
    return p

def deploy(opts, args):
    um = UserManager()
    mm = MessageManager()
    fm = Spooler(um, mm)
    fm.opts = opts
    fm.args = args
    fsm = fm.fork(None, None, None, Listener, [os.getpid()], {})
    fsm()

def main(opts, args):
    deploy(opts, args)

    return 0

if __name__ == "__main__":
    openlog("sprinkle", syslog.LOG_PERROR | syslog.LOG_PID)
    p = parser()
    opts, args = p.parse_args()
    rv = main(opts, args)
    closelog()
    sys.exit(rv)
